# coding=utf-8
# Filename: aanet.py
# pylint: disable=locally-disabled
"""
Pump for the Aanet data format.

"""
from __future__ import division, absolute_import, print_function

from km3pipe import Pump
from km3pipe.logger import logging
import os.path

log = logging.getLogger(__name__)  # pylint: disable=C0103


class AanetPump(Pump):
    """A pump for binary Aanet files."""

    def __init__(self, **context):
        super(self.__class__, self).__init__(**context)

        self.filename = self.get('filename')
        self.filenames = self.get('filenames') or []
        self.indices = self.get('indices')

        if not self.filename and not self.filenames:
            raise ValueError("No filename(s) defined")

        if self.filename:
            if "[index]" in self.filename and self.indices:
                self._parse_filenames()
            else:
                self.filenames.append(self.filename)

        self.header = None
        self.blobs = self.blob_generator()

    def _parse_filenames(self):
        prefix, suffix = self.filename.split('[index]')
        self.filenames += [prefix + str(i) + suffix for i in self.indices]

    def get_blob(self, index):
        NotImplementedError("Aanet currently does not support indexing.")

    def blob_generator(self):
        """Create a blob generator."""
        # pylint: disable:F0401,W0612
        import aa  # noqa
        from ROOT import EventFile

        for filename in self.filenames:
            print("Reading from file: {0}".format(filename))
            if not os.path.exists(filename):
                log.warn(filename + " not available: continue without it")
                continue
            event_file = EventFile(filename)
            self.header = event_file.rootfile().Get("Header")
            for event in event_file:
                blob = {'Evt': event,
                        'Hits': event.hits,
                        'MCHits': event.mc_hits,
                        'Tracks': event.trks,
                        'MCTracks': event.mc_trks,
                        'filename': filename,
                        'Header': self.header}
                yield blob
            del event_file

    def process(self, blob):
        return next(self.blobs)



# coding=utf-8
# Filename: db.py
# pylint: disable=locally-disabled
"""
Database utilities.

"""
from __future__ import division, absolute_import, print_function

from datetime import datetime
import ssl
import sys
import json

import pandas as pd

from km3pipe.tools import Timer
from km3pipe.config import Config
from km3pipe.logger import logging

if sys.version_info.major > 2:
    from urllib.parse import urlencode
    from urllib.request import (Request, build_opener,
                                HTTPCookieProcessor, HTTPHandler)
    from io import StringIO
    from http.cookiejar import CookieJar
else:
    from urllib import urlencode
    from urllib2 import (Request, build_opener,
                         HTTPCookieProcessor, HTTPHandler)
    from StringIO import StringIO
    from cookielib import CookieJar

__author__ = 'tamasgal'

log = logging.getLogger(__name__)  # pylint: disable=C0103


# Ignore invalid certificate error
try:
    ssl._create_default_https_context = ssl._create_unverified_context
except AttributeError:
    log.warn("Your SSL support is outdate. "
             "Please update your Python installation!")

LOGIN_URL = 'https://km3netdbweb.in2p3.fr/home.htm'
BASE_URL = 'https://km3netdbweb.in2p3.fr'


class DBManager(object):
    """A wrapper for the KM3NeT Web DB"""
    def __init__(self, username=None, password=None):
        "Create database connection"
        self.cookies = []
        self._parameters = None
        self._doms = None
        self._detectors = None
        self._opener = None
        if username is None:
            config = Config()
            username, password = config.db_credentials
        self.login(username, password)

    def datalog(self, parameter, run, maxrun=None, detid='D_ARCA001'):
        "Retrieve datalogs for given parameter, run(s) and detector"
        parameter = parameter.lower()
        if maxrun is None:
            maxrun = run
        with Timer('Database lookup'):
            return self._datalog(parameter, run, maxrun, detid)

    def _datalog(self, parameter, run, maxrun, detid):
        "Extract data from database"
        values = {'parameter_name': parameter,
                  'minrun': run,
                  'maxrun': maxrun,
                  'detid': detid,
                  }
        data = urlencode(values)
        content = self._get_content('streamds/datalognumbers.txt?' + data)
        if content.startswith('ERROR'):
            log.error(content)
            return None
        try:
            dataframe = pd.read_csv(StringIO(content), sep="\t")
        except ValueError:
            log.warning("Empty dataset")  # ...probably. Waiting for more info
            return pd.DataFrame()
        else:
            def convert_data(timestamp):
                return datetime.fromtimestamp(float(timestamp) / 1e3)
            dataframe['DATETIME'] = dataframe['UNIXTIME'].apply(convert_data)
            convert_unit = self.parameters.get_converter(parameter)
            dataframe['VALUE'] = dataframe['DATA_VALUE'].apply(convert_unit)
            dataframe.unit = self.parameters.unit(parameter)
            return dataframe

    def run_table(self, detid='D_ARCA001'):
        url = 'streamds/runs.txt?detid={0}'.format(detid)
        content = self._get_content(url)
        try:
            dataframe = pd.read_csv(StringIO(content), sep="\t")
        except ValueError:
            log.warning("Empty dataset")
            return None
        else:
            def convert_data(timestamp):
                return datetime.fromtimestamp(float(timestamp) / 1e3)
            converted = dataframe['UNIXSTARTTIME'].apply(convert_data)
            dataframe['DATETIME'] = converted
            return dataframe

    @property
    def detectors(self):
        if not self._detectors:
            self._detectors = self._get_detectors()
        return self._detectors

    def _get_detectors(self):
        content = self._get_content('streamds/detectors.txt')
        try:
            dataframe = pd.read_csv(StringIO(content), sep="\t")
        except ValueError:
            log.warning("Empty dataset")
            return pd.DataFrame()
        else:
            return dataframe

    @property
    def parameters(self):
        "Return the parameters container for quick access to their details"
        if self._parameters is None:
            self._load_parameters()
        return self._parameters

    def _load_parameters(self):
        "Retrieve a list of available parameters from the database"
        parameters = self._get_json('allparam/s')
        data = {}
        for parameter in parameters:  # There is a case-chaos in the DB
            data[parameter['Name'].lower()] = parameter
        self._parameters = ParametersContainer(data)

    @property
    def doms(self):
        if self._doms is None:
            self._load_doms()
        return self._doms

    def _load_doms(self):
        "Retrieve DOM information from the database"
        doms = self._get_json('domclbupiid/s')
        self._doms = DOMContainer(doms)

    def _get_json(self, url):
        "Get JSON-type content"
        content = self._get_content('jsonds/' + url)
        json_content = json.loads(content.decode())
        if json_content['Comment']:
            log.warn(json_content['Comment'])
        if json_content['Result'] != 'OK':
            raise ValueError('Error while retrieving the parameter list.')
        return json_content['Data']

    def _get_content(self, url):
        "Get HTML content"
        target_url = BASE_URL + '/' + url
        f = self.opener.open(target_url)
        content = f.read()
        return content

    @property
    def opener(self):
        "A reusable connection manager"
        if self._opener is None:
            opener = build_opener()
            for cookie in self.cookies:
                cookie_str = cookie.name + '=' + cookie.value
                opener.addheaders.append(('Cookie', cookie_str))
            self._opener = opener
        return self._opener

    def login(self, username, password):
        "Login to the databse and store cookies for upcoming requests."
        cj = CookieJar()
        opener = build_opener(HTTPCookieProcessor(cj), HTTPHandler())
        values = {'usr': username, 'pwd': password}
        data = urlencode(values)
        req = Request(LOGIN_URL, data.encode('utf-8'))
        f = opener.open(req)
        html = f.read()
        if 'Bad username or password' in str(html):
            log.error("Bad username or password!")
        self.cookies = cj


class ParametersContainer(object):
    """Provides easy access to parameters"""
    def __init__(self, parameters):
        self._parameters = parameters
        self._converters = {}

    @property
    def names(self):
        "A list of parameter names"
        return self._parameters.keys()

    def get_parameter(self, parameter):
        "Return a dict of given parameter"
        return self._parameters[parameter]

    def get_converter(self, parameter):
        """Generate unit conversion function for given parameter"""
        if parameter not in self._converters:
            param = self.get_parameter(parameter)
            try:
                scale = float(param['Scale'])
            except KeyError:
                scale = 1

            def convert(value):
                # easy_scale = float(param['EasyScale'])
                # easy_scale_multiplier = float(param['EasyScaleMultiplier'])
                return value * scale

            return convert

    def unit(self, parameter):
        "Get the unit for given parameter"
        return self._parameters[parameter.lower()]['Unit']


class DOMContainer(object):
    """Provides easy access to DOM parameters"""
    def __init__(self, doms):
        self._json = doms
        self._ids = []

    def ids(self, det_id):
        """Return a list of DOM IDs for given detector"""
        return [dom['DOMId'] for dom in self._json if dom['DetOID'] == det_id]

    def clbupi2domid(self, clb_upi, det_id):
        """Return DOM ID for given CLB UPI and detector"""
        lookup = [dom['DOMId'] for dom in self._json if
                  dom['DetOID'] == det_id and
                  dom['CLBUPI'] == clb_upi]
        if len(lookup) > 1:
            log.warn("Multiple entries found: {0}".format(lookup) + "\n" +
                     "Return the first one.")
        return lookup[0]



# coding=utf-8
# Filename: config.py
# pylint: disable=locally-disabled
"""
Tools for global configuration.

"""
from __future__ import division, absolute_import, print_function

import os
import stat
from six.moves.configparser import ConfigParser, NoSectionError
import getpass
try:
    input = raw_input
except NameError:
    pass

from km3pipe.logger import logging

__author__ = 'tamasgal'

log = logging.getLogger(__name__)  # pylint: disable=C0103

CONFIG_PATH = os.path.expanduser('~/.km3net')


class Config(object):
    def __init__(self):
        """Configuration manager for KM3NeT stuff"""
        self.config = ConfigParser()
        try:
            self._read_configuration()
        except IOError:
            log.warn("No configuration found at '{0}'".format(CONFIG_PATH))
        else:
            self._check_config_file_permissions()

    def _check_config_file_permissions(self):
        """Make sure that the configuration file is 0600"""
        allowed_modes = ['0600', '0o600']
        if oct(stat.S_IMODE(os.lstat(CONFIG_PATH).st_mode)) in allowed_modes:
            return True
        else:
            log.critical("Your config file is readable to others!\n" +
                         "Please execute `chmod 0600 {0}`".format(CONFIG_PATH))
            return False

    def _read_configuration(self):
        """Parse configuration file"""
        self.config.readfp(open(CONFIG_PATH))

    @property
    def db_credentials(self):
        """Return username and password for the KM3NeT WebDB."""
        try:
            username = self.config.get('DB', 'username')
            password = self.config.get('DB', 'password')
        except NoSectionError:
            username = input("Please enter your KM3NeT DB username: ")
            password = getpass.getpass("Password: ")
        return username, password



# coding=utf-8
# Filename: pickle.py
# pylint: disable=locally-disabled
"""
Pump for the pickle data format.

"""
from __future__ import division, absolute_import, print_function

import pickle
from collections import namedtuple

from km3pipe import Pump
from km3pipe.logger import logging


log = logging.getLogger(__name__)  # pylint: disable=C0103


class PicklePump(Pump):
    """A pump for Python pickles."""
    def __init__(self, **context):
        super(self.__class__, self).__init__(**context)

        self.filename = self.get('filename')
        self.events = pickle.load(open(self.filename, "rb"))

    def get_blob(self, index):
        hits = self.events[0]
        t0 = min([hit[3] for hit in hits])
        raw_hits = []
        for hit in hits:
            raw_hit = RawHit(hit[0], hit[1], hit[2], hit[3] - t0)
            print(raw_hit)
            raw_hits.append(raw_hit)
        blob = {'EvtRawHits': raw_hits}
        return blob


RawHit = namedtuple('RawHit', 'id pmt_id tot time')
